_VERSION = 0.9

header_comment = """--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--
--
-- Multitap Description Table
--
-- Generate this file by using the accompanying Python script.
--      python GenerateMultitaps.py "C:\\path\\to\\simfile.sm"
--      GenerateMultitaps(r"C:\\path\\to\\simfile.sm")
--
-- The script uses the 2P side of the double chart in the matching slot
-- to group tap notes on the 1P side into multitap regions.
-- By copying the 4-panel chart to the corresponding 8-panel slot and
-- inserting holds in the corresponding lane on the other side, you can
-- tell the script which swaths of notes should become multitaps.
--
--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--[[##]]--

"""

from dataclasses import dataclass, field
import os
import sys
import shutil

try:
    import simfile
    from simfile.notes import NoteData, NoteType
    from simfile.notes.group import group_notes
except Exception as e:
    print("Please install the simfile package. https://simfile.readthedocs.io/en/latest/\n>   pip3 install simfile")
    sys.exit(1)

@dataclass
class Multitap:
    lane: int                                           # Column in which multitap appears
    taps: 'typing.List' = field(default_factory=list)   # Beats at which multitap should fire
    peak: 'typing.Any' = None                           # Manual elasticity parameter

    def to_lua(self, indent='\t\t'):
        s = ''
        s += f'{indent}{{lane = {self.lane + 1}, taps = {{\n'   # Lua is one-indexed
        for t in self.taps:
            s += f'{indent}\t{t},\n'
        if self.peak is not None:
            s += f'{indent}}}, peak = {self.peak:0.3f}}},\n'
        else:
            s += f'{indent}}}}},\n'
        return s

def IdentifyMultitaps(note_data):
    mt_groups = []

    # Look for holds in the P2 side of the double chart.
    for mt_region in group_notes(
        note_data,
        include_note_types={NoteType.HOLD_HEAD, NoteType.ROLL_HEAD, NoteType.TAIL},
        join_heads_to_tails=True
    ):
        if mt_region[0].column in [4, 5, 6, 7]:
            mt_desc = Multitap(lane=mt_region[0].column - 4)
            for n in note_data:
                # Look for notes that are in the corresponding P1 column,
                # and that occur between the hold head and hold tail (inclusive).
                if n.column == mt_region[0].column - 4 and \
                   n.beat >= mt_region[0].beat and \
                   n.beat <= mt_region[0].tail_beat:
                    mt_desc.taps.append(n.beat)
            mt_groups.append(mt_desc)

    return mt_groups


def IdentifyMultitapsAll(sim):
    mt_data = {}
    for c in sim.charts:
        if c.stepstype.lower() == 'dance-double':
            slot = c.difficulty
            mt_groups = IdentifyMultitaps(NoteData(c))
            mt_data[slot] = mt_groups
    return mt_data


def GenerateMultitaps(path_to_simfile):
    # Assumes the FG change lives at <simfile dir>/multitap.
    path_to_multitap_data = os.path.join(os.path.dirname(path_to_simfile), 'multitap', 'multitap_data.lua')
    if os.path.exists(path_to_multitap_data):
        # Save aside a copy of the old multitap data, just in case
        shutil.copy2(path_to_multitap_data, path_to_multitap_data + '.old')

    sim = simfile.open(path_to_simfile)
    mt_data = IdentifyMultitapsAll(sim)

    with open(path_to_multitap_data, 'w') as fp:
        fp.write(header_comment)
        fp.write('multitaps = {\n')
        fp.write(f'\t_version = {{{int(_VERSION)},{int(_VERSION*10) - int(_VERSION)*10}}},\n')
        for mt_slot, mt_groups in mt_data.items():
            fp.write(f'\t{mt_slot} = {{\n')
            for mt_desc in mt_groups:
                fp.write(mt_desc.to_lua())
            fp.write('\t},\n')
        fp.write('}\n')


if __name__ == '__main__':
    GenerateMultitaps(sys.argv[1])
